"""Module to check the library structure."""

from vidmux.library_structure.core import (
    IssueCode,
    ValidationIssue,
    Severity,
    registry,
    run_validation,
)
from vidmux.library_structure.structure_scan import scan_library_structure


def load_default_rules() -> None:
    """Load and register the default rules."""
    from vidmux.library_structure import rules


__all__ = [
    "IssueCode",
    "ValidationIssue",
    "Severity",
    "load_default_rules",
    "registry",
    "run_validation",
    "scan_library_structure",
]



"""Collection of CLI programs."""

import argparse
from pathlib import Path

import vidmux.srt_tools as srt_tools
from vidmux.language_tools import set_languages
from vidmux.library_structure import scan_library_structure
from vidmux.mux_audio_tracks import mux_audio_tracks
from vidmux.video_inspection import get_video_resolution
from vidmux.video_library_scan import scan_mode


def get_library_structur_parser(
    subparsers: argparse._SubParsersAction | None = None,
    prog: str | None = None,
    formatter_class: type | None = None,
) -> argparse.ArgumentParser:
    """
    Create and configure the argument parser for 'lib-structure'.

    This function can either add a subparser to an existing ArgumentParser
    (via `subparsers`) or create a standalone parser when called independently.
    Useful for modular CLI designs.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction, optional
        Subparsers object from the main parser to which this parser should be added.
        If None, a standalone ArgumentParser is created instead.
    prog : str, optional
        The program name used in standalone mode. Ignored if `subparsers` is provided.
    formatter_class : type, optional
        The formatter class to be used for argument help formatting. Defaults to
        argparse.ArgumentDefaultsHelpFormatter.

    Returns
    -------
    argparse.ArgumentParser
        The configured argument parser (necessary esp. for standalone mode).
    """
    parser_options = {
        "description": "Scan the structure of a library.",
        "formatter_class": formatter_class or argparse.ArgumentDefaultsHelpFormatter,
    }
    if subparsers:
        parser = subparsers.add_parser(
            "lib-structure", help=parser_options["description"], **parser_options
        )
    else:
        parser = argparse.ArgumentParser(prog=prog, **parser_options)

    parser.add_argument(
        "library",
        type=Path,
        help="Path to the library directory.",
    )
    parser.add_argument(
        "--extensions",
        metavar="EXTENSION",
        nargs="+",
        default=[".mp4", ".mkv", ".avi", ".mov", ".wmv", ".flv", ".webm"],
        help="File extensions to include.",
    )
    parser.add_argument(
        "--print",
        dest="show",
        action="store_true",
        help="Print results to the console.",
    )
    parser.add_argument(
        "--json", metavar="FILE", type=Path, help="Path to output JSON file."
    )
    parser.add_argument(
        "--csv", metavar="FILE", type=Path, help="Path to output CSV file."
    )

    return parser


def get_mux_audio_tracks_parser(
    subparsers: argparse._SubParsersAction | None = None,
    prog: str | None = None,
    formatter_class: type | None = None,
) -> argparse.ArgumentParser:
    """
    Create and configure the argument parser for 'mux-audio-tracks'.

    This function can either add a subparser to an existing ArgumentParser
    (via `subparsers`) or create a standalone parser when called independently.
    Useful for modular CLI designs.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction, optional
        Subparsers object from the main parser to which this parser should be added.
        If None, a standalone ArgumentParser is created instead.
    prog : str, optional
        The program name used in standalone mode. Ignored if `subparsers` is provided.
    formatter_class : type, optional
        The formatter class to be used for argument help formatting. Defaults to
        argparse.ArgumentDefaultsHelpFormatter.

    Returns
    -------
    argparse.ArgumentParser
        The configured argument parser (necessary esp. for standalone mode).
    """
    parser_options = {
        "description": "Mux multiple audio tracks into one video file using FFmpeg.",
        "formatter_class": formatter_class or argparse.ArgumentDefaultsHelpFormatter,
    }

    if subparsers:
        parser = subparsers.add_parser(
            "mux-audio-tracks", help=parser_options["description"], **parser_options
        )
    else:
        parser = argparse.ArgumentParser(prog=prog, **parser_options)

    parser.add_argument(
        "output",
        type=Path,
        help="Output file (e.g. output.mp4)",
    )

    parser.add_argument(
        "inputs",
        nargs="+",
        help=(
            "Pairs or triplets of inputs: input_file lang [title]. "
            "Example: video_en.mp4 eng 'English Audio' video_de.mp4 deu 'Deutsch'"
        ),
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Only print ffmpeg command without executing.",
    )

    return parser


def get_resolution_parser(
    subparsers: argparse._SubParsersAction | None = None,
    prog: str | None = None,
    formatter_class: type | None = None,
) -> argparse.ArgumentParser:
    """
    Create and configure the argument parser for 'srt-tools'.

    This function can either add a subparser to an existing ArgumentParser
    (via `subparsers`) or create a standalone parser when called independently.
    Useful for modular CLI designs.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction, optional
        Subparsers object from the main parser to which this parser should be added.
        If None, a standalone ArgumentParser is created instead.
    prog : str, optional
        The program name used in standalone mode. Ignored if `subparsers` is provided.
    formatter_class : type, optional
        The formatter class to be used for argument help formatting. Defaults to
        argparse.ArgumentDefaultsHelpFormatter.

    Returns
    -------
    argparse.ArgumentParser
        The configured argument parser (necessary esp. for standalone mode).
    """
    parser_options = {
        "description": "Get the video resolution for a file or URL.",
        "formatter_class": formatter_class or argparse.ArgumentDefaultsHelpFormatter,
    }
    if subparsers:
        parser = subparsers.add_parser(
            "resolution", help=parser_options["description"], **parser_options
        )
    else:
        parser = argparse.ArgumentParser(prog=prog, **parser_options)

    parser.add_argument(
        "path_or_url",
        type=str,
        help="Path or URL to the video.",
    )

    return parser


def get_scan_library_parser(
    subparsers: argparse._SubParsersAction | None = None,
    prog: str | None = None,
    formatter_class: type | None = None,
) -> argparse.ArgumentParser:
    """
    Create and configure the argument parser for 'scan'.

    This function can either add a subparser to an existing ArgumentParser
    (via `subparsers`) or create a standalone parser when called independently.
    Useful for modular CLI designs.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction, optional
        Subparsers object from the main parser to which this parser should be added.
        If None, a standalone ArgumentParser is created instead.
    prog : str, optional
        The program name used in standalone mode. Ignored if `subparsers` is provided.
    formatter_class : type, optional
        The formatter class to be used for argument help formatting. Defaults to
        argparse.ArgumentDefaultsHelpFormatter.

    Returns
    -------
    argparse.ArgumentParser
        The configured argument parser (necessary esp. for standalone mode).
    """
    parser_options = {
        "description": "Scan videos of a library, e.g. for audio and subtitle tracks.",
        "formatter_class": formatter_class or argparse.ArgumentDefaultsHelpFormatter,
    }
    if subparsers:
        parser = subparsers.add_parser(
            "scan", help=parser_options["description"], **parser_options
        )
    else:
        parser = argparse.ArgumentParser(prog=prog, **parser_options)

    parser.add_argument(
        "library",
        type=Path,
        help="Path to the video library directory.",
    )
    parser.add_argument(
        "--extensions",
        metavar="EXTENSION",
        nargs="+",
        default=[".mp4", ".mkv", ".avi", ".mov"],
        help="File extensions to include.",
    )
    parser.add_argument(
        "--print",
        dest="show",
        action="store_true",
        help="Print results to the console.",
    )
    parser.add_argument(
        "--json", metavar="FILE", type=Path, help="Path to output JSON file."
    )
    parser.add_argument(
        "--csv", metavar="FILE", type=Path, help="Path to output CSV file."
    )

    return parser


def get_set_language_parser(
    subparsers: argparse._SubParsersAction | None = None,
    prog: str | None = None,
    formatter_class: type | None = None,
) -> argparse.ArgumentParser:
    """
    Create and configure the argument parser for 'set-language'.

    This function can either add a subparser to an existing ArgumentParser
    (via `subparsers`) or create a standalone parser when called independently.
    Useful for modular CLI designs.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction, optional
        Subparsers object from the main parser to which this parser should be added.
        If None, a standalone ArgumentParser is created instead.
    prog : str, optional
        The program name used in standalone mode. Ignored if `subparsers` is provided.
    formatter_class : type, optional
        The formatter class to be used for argument help formatting. Defaults to
        argparse.ArgumentDefaultsHelpFormatter.

    Returns
    -------
    argparse.ArgumentParser
        The configured argument parser (necessary esp. for standalone mode).
    """
    parser_options = {
        "description": "Set audio and subtitle language metadata.",
        "formatter_class": formatter_class or argparse.ArgumentDefaultsHelpFormatter,
    }
    if subparsers:
        parser = subparsers.add_parser(
            "set-language", help=parser_options["description"], **parser_options
        )
    else:
        parser = argparse.ArgumentParser(prog=prog, **parser_options)

    parser.add_argument("input_file", type=Path, help="Input video file.")
    parser.add_argument("output_file", type=Path, help="Output video file.")
    parser.add_argument(
        "--audio-lang",
        metavar="LANGUAGE",
        nargs="+",
        required=True,
        help="Audio languages (e.g., deu eng).",
    )
    parser.add_argument(
        "--subtitle-lang",
        metavar="LANGUAGE",
        nargs="*",
        default=[],
        help="Subtitle languages (optional).",
    )
    parser.add_argument(
        "--audio-title",
        metavar="TITLE",
        nargs="*",
        help="Optional titles for audio tracks.",
    )
    parser.add_argument(
        "--subtitle-title",
        metavar="TITLE",
        nargs="*",
        help="Optional titles for subtitle tracks.",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Only print ffmpeg command without executing.",
    )

    return parser


def get_srt_tool_parser(
    subparsers: argparse._SubParsersAction | None = None,
    prog: str | None = None,
    formatter_class: type | None = None,
) -> argparse.ArgumentParser:
    """
    Create and configure the argument parser for 'srt-tools'.

    This function can either add a subparser to an existing ArgumentParser
    (via `subparsers`) or create a standalone parser when called independently.
    Useful for modular CLI designs.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction, optional
        Subparsers object from the main parser to which this parser should be added.
        If None, a standalone ArgumentParser is created instead.
    prog : str, optional
        The program name used in standalone mode. Ignored if `subparsers` is provided.
    formatter_class : type, optional
        The formatter class to be used for argument help formatting. Defaults to
        argparse.ArgumentDefaultsHelpFormatter.

    Returns
    -------
    argparse.ArgumentParser
        The configured argument parser (necessary esp. for standalone mode).
    """
    parser_options = {
        "description": "Shift timestamps of a SRT file.",
        "formatter_class": formatter_class or argparse.ArgumentDefaultsHelpFormatter,
    }
    if subparsers:
        parser = subparsers.add_parser(
            "srt-tools", help=parser_options["description"], **parser_options
        )
    else:
        parser = argparse.ArgumentParser(prog=prog, **parser_options)

    parser.add_argument("input_file", help="Original SRT file.")
    parser.add_argument(
        "-s",
        "--shift",
        metavar="SECONDS",
        type=float,
        required=True,
        help="Timeshift in seconds (e.g. 1.5 or -0.8).",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        dest="output_file",
        help="Output SRT file (if not provided: stdout or --inplace).",
    )
    parser.add_argument(
        "--inplace",
        action="store_true",
        help="Overwrite original file (a backup will be created).",
    )
    parser.add_argument(
        "--show-count",
        action="store_true",
        help="Show number of changed timestamps.",
    )

    return parser


def main() -> None:
    """Run main CLI programm."""
    formatter_class = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        prog="vidmux",
        description="Inspect and modify video/audio/subtitle tracks using FFmpeg.",
        formatter_class=formatter_class,
    )
    subparsers = parser.add_subparsers(dest="feature")

    feature_parsers = (
        get_library_structur_parser,
        get_mux_audio_tracks_parser,
        get_resolution_parser,
        get_scan_library_parser,
        get_set_language_parser,
        get_srt_tool_parser,
    )
    for feature_parser in feature_parsers:
        feature_parser(subparsers, formatter_class=formatter_class)

    args = parser.parse_args()
    match args.feature:
        case "lib-structure":
            scan_library_structure(
                args.library,
                args.extensions,
                show=args.show,
                json_file=args.json,
                csv_file=args.csv,
            )
        case "mux-audio-tracks":
            mux_audio_tracks(args.output, args.inputs, dry_run=args.dry_run)
        case "resolution":
            resolution = get_video_resolution(args.path_or_url)
            print(f"Video resolution of '{args.path_or_url}':\n{resolution}")
        case "scan":
            scan_mode(
                args.library,
                args.extensions,
                show=args.show,
                json_file=args.json,
                csv_file=args.csv,
            )
        case "set-language":
            set_languages(
                input_file=args.input_file,
                output_file=args.output_file,
                audio_languages=args.audio_lang,
                subtitle_languages=args.subtitle_lang,
                audio_titles=args.audio_title,
                subtitle_titles=args.subtitle_title,
                dry_run=args.dry_run,
            )
        case "srt-tools":
            srt_tools.process_file(
                args.input_file,
                args.shift,
                inplace=args.inplace,
                output_file=args.output_file,
                show_count=args.show_count,
            )
        case _:
            parser.print_help()
            parser.exit(message="Run again and specify a supported command.")



"""Test for muxing audio tracks."""

from pathlib import Path
from unittest.mock import patch

import pytest

from vidmux.mux_audio_tracks import mux_audio_tracks


def test_mux_fails_with_missing_args(tmp_path: Path) -> None:
    """Test failing on purpose."""
    outfile = tmp_path / "out.mp4"
    with pytest.raises(SystemExit):
        mux_audio_tracks(outfile, ["only_one_arg"])


def test_mux_command(tmp_path: Path) -> None:
    """Test whether the mux command is constructed correctly without running ffmpeg."""
    outfile = tmp_path / "outfile.mp4"

    infile_1 = tmp_path / "video_1.mp4"
    infile_2 = tmp_path / "video_2.mp4"
    infile_3 = tmp_path / "video_3.mp4"

    # Create empty dummy input files
    for file in (infile_1, infile_2, infile_3):
        file.touch()

    # Patch subprocess.run so ffmpeg is not actually executed
    with patch("subprocess.run") as mock_run:
        mux_audio_tracks(
            outfile,
            [
                infile_1,
                "deu",
                infile_2,
                "deu",
                "Deutscher Kommentar",
                infile_3,
                "eng",
            ],
        )

    # Extract the command that would have been run
    args = mock_run.call_args[0][0]

    # Define expected ffmpeg arguments
    expected_args = [
        "ffmpeg",
        "-i",
        str(infile_1),
        "-i",
        str(infile_2),
        "-i",
        str(infile_3),
        # mapping: first input has video+audio, others only audio
        "-map",
        "0:v:0",
        "-map",
        "0:a:0",
        "-map",
        "1:a:0",
        "-map",
        "2:a:0",
        "-c:v",
        "copy",
        "-c:a",
        "copy",
        # metadata
        "-metadata:s:a:0",
        "language=deu",
        "-metadata:s:a:1",
        "language=deu",
        "-metadata:s:a:1",
        "title=Deutscher Kommentar",
        "-metadata:s:a:2",
        "language=eng",
        str(outfile),
    ]

    # Check whether expected and true arguments are equal
    assert args == expected_args, (
        "FFmpeg arguments do not match expected command.\n"
        f"Expected:\n  {expected_args}\nGot:\n  {args}"
    )

    # Ensure only one subprocess call
    mock_run.assert_called_once_with(expected_args, check=True)


def test_mux_command_with_fixture(tmp_path: Path, ffmpeg_mock) -> None:
    """Test whether the mux command is constructed correctly without running ffmpeg."""
    outfile = tmp_path / "outfile.mp4"

    infile_1 = tmp_path / "video_1.mp4"
    infile_2 = tmp_path / "video_2.mp4"
    infile_3 = tmp_path / "video_3.mp4"

    # Create empty dummy input files
    for file in (infile_1, infile_2, infile_3):
        file.touch()

    mux_audio_tracks(
        outfile,
        [
            infile_1,
            "deu",
            infile_2,
            "deu",
            "Deutscher Kommentar",
            infile_3,
            "eng",
        ],
    )

    # Define expected ffmpeg arguments
    expected_args = [
        "ffmpeg",
        "-i",
        str(infile_1),
        "-i",
        str(infile_2),
        "-i",
        str(infile_3),
        # mapping: first input has video+audio, others only audio
        "-map",
        "0:v:0",
        "-map",
        "0:a:0",
        "-map",
        "1:a:0",
        "-map",
        "2:a:0",
        "-c:v",
        "copy",
        "-c:a",
        "copy",
        # metadata
        "-metadata:s:a:0",
        "language=deu",
        "-metadata:s:a:1",
        "language=deu",
        "-metadata:s:a:1",
        "title=Deutscher Kommentar",
        "-metadata:s:a:2",
        "language=eng",
        str(outfile),
    ]

    ffmpeg_mock.assert_called_with(expected_args)


@pytest.mark.parametrize(
    "inputs",
    [
        ["a.mp4", "eng"],
        ["a.mp4", "eng", "English"],
        ["a.mp4", "eng", "English", "b.mp4", "deu", "Deutsch"],
    ],
)
def test_mux_argument_patterns(tmp_path: Path, inputs: list[str]) -> None:
    """Test multiple argument patterns."""
    outfile = tmp_path / "out.mp4"
    for name in inputs[::3]:
        (tmp_path / name).touch()
    with patch("subprocess.run") as mock_run:
        mux_audio_tracks(
            outfile, [tmp_path / x if x.endswith(".mp4") else x for x in inputs]
        )
    mock_run.assert_called_once()



"""Define the high-level API."""

import csv
import json
from pathlib import Path

from vidmux.library_structure.core import run_validation
from vidmux.library_structure import rules


def save_json(results: list[dict], path: Path) -> None:
    """Save results to a JSON file."""
    with open(path, "w", encoding="utf-8") as file:
        json.dump(results, file, indent=2, ensure_ascii=False)

    print(f"Saved JSON: {path}")


def save_csv(results: list[dict], path: Path) -> None:
    """Save results to a CSV file."""
    with open(path, "w", newline="", encoding="utf-8") as file:
        writer = csv.writer(file)
        writer.writerow(["filename", "type", "code", "description", "message"])
        for report in results:
            for issue in report["issues"]:
                writer.writerow(
                    [
                        issue["path"],
                        issue["severity"],
                        issue["code"],
                        issue["description"],
                        issue["message"],
                    ]
                )

    print(f"Save CSV: {path}")


def make_issue_string(issue: dict) -> str:
    """Convert an issue to a string."""
    return (
        f"[{issue['severity']}] {issue['code']} ({issue['description']}): "
        f"{issue['message']}"
    )


def print_to_terminal(results: list[dict]) -> None:
    """Print results to terminal."""
    print("Issues in the library:")
    for report in results:
        if report["issues"]:
            msg = f"Issues for {report['path']}:\n\t"
            msg = msg + "\n\t".join(
                make_issue_string(issue) for issue in report["issues"]
            )
            print(msg)


def scan_library_structure(
    library: Path,
    extensions: list[str],
    show: bool = True,
    json_file: Path | None = None,
    csv_file: Path | None = None,
) -> bool:
    """Run the scan and save/show the output."""
    if not (show or json_file or csv_file):
        print("No output specified. Use --print, --json or --csv.")
        return False

    files = [file for file in library.rglob("*") if file.suffix.lower() in extensions]
    result = run_validation(files)

    if show:
        print_to_terminal(result)

    if json_file:
        save_json(result, json_file)

    if csv_file:
        save_csv(result, csv_file)

    return True



"""Tools to mux audio tracks."""

import subprocess
import shlex
import sys
from pathlib import Path


def mux_audio_tracks(output_file, inputs, dry_run: bool = False) -> None:
    """
    Execute the 'mux-audio-tracks' command using FFmpeg.
    """
    if len(inputs) < 2:
        sys.exit("Error: Need at least one input file and language code.")

    input_args = []
    map_args = []
    metadata_args = []

    cmd_idx = 0
    input_idx = 0
    while input_idx < len(inputs):
        if input_idx + 1 >= len(inputs):
            sys.exit("Error: Each input requires at least a file and language code.")

        input_file = str(inputs[input_idx])
        lang = inputs[input_idx + 1]
        title = None

        # Optional title if the next argument is not a file path that exists
        if input_idx + 2 < len(inputs) and not Path(inputs[input_idx + 2]).exists():
            title = inputs[input_idx + 2]
            input_idx += 3
        else:
            input_idx += 2

        if cmd_idx == 0:
            input_args += ["-i", input_file]
            map_args += ["-map", "0:v:0", "-map", "0:a:0"]
        else:
            input_args += ["-i", input_file]
            map_args += ["-map", f"{cmd_idx}:a:0"]

        metadata_args += ["-metadata:s:a:%d" % cmd_idx, f"language={lang}"]
        if title:
            metadata_args += ["-metadata:s:a:%d" % cmd_idx, f"title={title}"]

        cmd_idx += 1

    command = [
        "ffmpeg",
        *input_args,
        *map_args,
        "-c:v",
        "copy",
        "-c:a",
        "copy",
        *metadata_args,
        str(output_file),
    ]

    print("FFmpeg command:")
    print(" ".join(shlex.quote(part) for part in command))

    if not dry_run:
        subprocess.run(command, check=True)



"""Main entry point for vidmux."""

import shutil
import sys

from vidmux.cli import main as cli_mode


def ensure_ffmpeg() -> None:
    """Exit if 'ffmpeg' is not found."""
    if not shutil.which("ffmpeg"):
        sys.exit("Error: FFmpeg not found. Please install it and ensure it is in PATH.")


def main():
    """Entry point for "vidmux"and "python -m vidmux"."""
    ensure_ffmpeg()
    cli_mode()


if __name__ == "__main__":
    main()



"""Test the library structure submodule."""

from pathlib import Path

import pytest

from vidmux.library_structure import registry, load_default_rules, run_validation

MOVIES = {
    "/Example Movie A.mp4": ["MISSING_YEAR"],
    "/Example Movie B (20).mp4": ["MISSING_YEAR"],
    "/Example Movie C (2009).mp4": [],
    "Example Movie D (2003)/Example Movie D (2003) [EN+DE].mp4": [],
    "Example Movie E (2003)/Example Movie E2 (2003).mp4": [
        "FILE_AND_FOLDER_NAME_DIFFER"
    ],
    "Example Movie F.mp4": ["MISSING_YEAR", "FILE_NOT_IN_FOLDER"],
    "Example Movie G (2003).mp4": ["FILE_NOT_IN_FOLDER"],
}


@pytest.fixture(scope="module")
def example_library(tmp_path_factory: pytest.TempPathFactory) -> Path:
    tmp_path = tmp_path_factory.mktemp("library")

    for name in MOVIES.keys():
        if "/" in name:
            # File should be in folder (name equals filename if no folder is provided)
            parts = name.split("/")
            folder_name = parts[0] if len(parts) == 1 else parts[-2]
            file_name = parts[-1]
            if not folder_name:
                folder_name = file_name.removesuffix(".mp4")
        else:
            # File should not be in a folder
            folder_name = None

        path = tmp_path / folder_name / file_name if folder_name else tmp_path / name

        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()

    return tmp_path


@pytest.fixture(autouse=True, scope="module")
def _load_default_rules():
    """Automatically load default rules for this test module."""
    load_default_rules()


@pytest.mark.parametrize("total_name,expected_codes", MOVIES.items())
def test_library_issues(example_library, total_name, expected_codes):
    """Test each movie against expected validation results."""
    reports = run_validation(example_library.rglob("*.mp4"))
    file_name = total_name.split("/")[-1]

    # Find report for file
    entry = next((report for report in reports if file_name in report["path"]), None)
    assert entry is not None, f"{file_name} missing in report"

    codes = [issue["code"] for issue in entry["issues"]]
    assert sorted(codes) == sorted(expected_codes)


# def test_debug_report(example_library):
#     """Show the report of the example library check and fail."""
#     import json

#     report = run_validation(example_library.rglob("*.mp4"))
#     print(json.dumps(report, indent=2))
#     print("Used rules:", registry.rules)
#     assert False, "Manual debug stop"



"""Define core features for the validation, esp. issues and validation results."""

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Callable, Any


class Severity(Enum):
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"


class IssueCode(Enum):
    FILE_NOT_IN_FOLDER = "File not in its own folder"
    FILE_AND_FOLDER_NAME_DIFFER = "Filename differs from folder"
    BAD_CHARACTER_IN_NAME = "Invalid character in filename"
    MISSING_YEAR = "Missing year in title"


@dataclass
class ValidationIssue:
    path: str
    code: IssueCode
    message: str
    severity: Severity

    def to_dict(self) -> dict:
        """Return a dictionary representation of the instance values."""
        return {
            "path": str(self.path),
            "severity": self.severity.name,
            "code": self.code.name,
            "description": self.code.value,
            "message": self.message,
        }


@dataclass
class CheckResult:
    path: Path
    issues: list[ValidationIssue] = field(default_factory=list)

    def add_issue(
        self, code: IssueCode, message: str, severity: Severity = Severity.ERROR
    ) -> None:
        """Add an issue to the issue list."""
        self.issues.append(ValidationIssue(self.path, code, message, severity))

    def to_dict(self) -> dict:
        """Return a dictionary representation of the instance values."""
        return {
            "path": str(self.path),
            "issues": [issue.to_dict() for issue in self.issues],
        }


class Rule:
    """Single rule."""

    def __init__(
        self,
        name: str,
        func: Callable[[Path, dict[str, Any]], list[ValidationIssue]],
        default_enabled: bool = True,
        default_severity: Severity | None = None,
        default_params: dict[str, Any] | None = None,
    ):
        self.name = name
        self.func = func
        self.default_enabled = default_enabled
        self.default_severity = default_severity
        self.default_params = default_params or {}

    def run(self, path: Path, config: dict[str, Any]) -> list[ValidationIssue]:
        """Run the check of the rule (if active)."""
        rule_config = config.get("rules", {}).get(self.name, {})
        enabled = rule_config.get("enabled", self.default_enabled)
        if not enabled:
            return []

        # Merge Parameters
        params = {**self.default_params, **rule_config.get("params", {})}
        severity_override = rule_config.get("severity", self.default_severity)

        issues = self.func(path, params)

        # Patch severity if necessary
        if severity_override:
            for issue in issues:
                issue.severity = severity_override

        return issues


class RuleRegistry:
    """Save registered rules for checking."""

    def __init__(self):
        self.rules: dict[str, Rule] = {}

    def register(
        self,
        name: str | None = None,
        default_enabled: bool = True,
        default_severity: Severity | None = None,
        default_params: dict[str, Any] | None = None,
    ):
        """Register a rule."""

        def decorator(func):
            """Allow decorator syntax to register rules."""
            rule_name = name or func.__name__
            self.rules[rule_name] = Rule(
                rule_name,
                func,
                default_enabled,
                default_severity,
                default_params,
            )

            return func

        return decorator

    def run_all(self, path: Path, config: dict[str, Any]) -> CheckResult:
        """Run a check of all registered rules."""
        result = CheckResult(path)
        for rule in self.rules.values():
            for issue in rule.run(path, config):
                result.issues.append(issue)

        return result


# Global registry instance
registry = RuleRegistry()


def run_validation(
    files: list[Path], config: dict[str, Any] | None = None
) -> list[dict]:
    """Return a check report list of dicts for a list of files."""
    config = config or {}
    results = []
    for file in files:
        result = registry.run_all(file, config)
        results.append(result.to_dict())

    return results



"""Tools for .srt files (subtitles)."""

import re
from pathlib import Path

# Regex for SRT timestamps: allows H or HH, allows "," or "." for ms
TIMESTAMP_RE = re.compile(
    r"(\d{1,2}):([0-5]\d):([0-5]\d)[\.,](\d{1,3})\s*-->\s*"
    r"(\d{1,2}):([0-5]\d):([0-5]\d)[\.,](\d{1,3})"
)


def timestamp_to_milliseconds(
    hours: str, minutes: str, seconds: str, milliseconds: str
) -> int:
    """Convert timestamp components to ms."""
    # Normalise ms to 3 digits
    ms_str = milliseconds.ljust(3, "0") if len(milliseconds) < 3 else milliseconds[:3]

    return (int(hours) * 3600 + int(minutes) * 60 + int(seconds)) * 1000 + int(ms_str)


def timestamp_from_milliseconds(total_ms: int) -> str:
    """Create timestamp from ms."""
    if total_ms < 0:
        total_ms = 0
    ms = total_ms % 1000
    total_s = total_ms // 1000
    s = total_s % 60
    total_m = total_s // 60
    m = total_m % 60
    h = total_m // 60

    return f"{h:02d}:{m:02d}:{s:02d},{ms:03d}"


def shift_line(re_match: re.Match, shift_ms: int) -> str:
    """Return time-shifted timestamp line."""
    h1, m1, s1, ms1, h2, m2, s2, ms2 = re_match.groups()
    start_ms = timestamp_to_milliseconds(h1, m1, s1, ms1)
    end_ms = timestamp_to_milliseconds(h2, m2, s2, ms2)
    new_start = timestamp_from_milliseconds(start_ms + shift_ms)
    new_end = timestamp_from_milliseconds(end_ms + shift_ms)

    return f"{new_start} --> {new_end}"


def process_text(text: str, shift_seconds: float) -> tuple[str, int]:
    """Shift all timestamps in the text."""
    shift_ms = int(round(shift_seconds * 1000))

    def repl(re_match: re.Match):
        """Return the shifted timestamp line."""
        return shift_line(re_match, shift_ms)

    new_text, count = TIMESTAMP_RE.subn(repl, text)

    return new_text, count


def process_file(
    input_file: str | Path,
    shift_seconds: float,
    inplace: bool = True,
    output_file: str | Path | None = None,
    show_count: bool = True,
):
    """Shift the timestamps of a file."""
    input_file = Path(input_file)
    if not input_file.exists():
        msg = f"'{input_file}' does not exist!"
        raise FileNotFoundError(msg)

    text = input_file.read_text(encoding="utf-8-sig")  # Read UTF-8 (+ potential BOM)
    new_text, count = process_text(text, shift_seconds)

    if inplace:
        # Create backup
        backup_file = input_file.with_suffix(input_file.suffix + ".bak")
        input_file.rename(backup_file)
        # Set output filename to input filename
        output_file = input_file
        output_file.write_text(new_text, encoding="utf-8")
        # Show number of changes
        if show_count:
            print(
                f"Wrote {output_file} (backup: {backup_file}), "
                f"changed timestamps: {count}"
            )
    else:
        if output_file:
            output_file = Path(output_file)
            output_file.write_text(new_text, encoding="utf-8")
            # Show number of changes
            if show_count:
                print(f"Wrote {output_file}, changed timestamps: {count}")
        else:
            # No outfile -> stdout
            print(new_text)
            # Show number of changes
            if show_count:
                print(f"\nChanged timestamps: {count}")



"""Tools to scan a video library."""

import csv
import json
from pathlib import Path

from vidmux.video_inspection import (
    get_audio_tracks,
    get_format,
    get_subtitles,
    get_video_tracks,
)


def make_track_entries(tracks: list) -> list:
    """Return the track list where each track has index, language, codec and title."""
    return [
        {
            "index": idx,
            "language": track.get("tags", {}).get("language", "unknown"),
            "codec": track.get("codec_name", "unknown"),
            "title": track.get("tags", {}).get("title", ""),
        }
        for idx, track in enumerate(tracks)
    ]


def show_track_entries(
    tracks: list, name: str = "track", output: callable = print
) -> None:
    """Show tracks."""
    for track in tracks:
        idx = track["index"]
        language = track["language"]
        codec = track["codec"]
        title = track["title"]
        output(f"\t{name.capitalize()} {idx+1}: {language=}, {codec=}, {title=}")


def scan_video_library(
    library_path: Path, extensions: list[str] | None = None
) -> list[dict]:
    """Scan all videos in 'library_path' recursively."""
    if extensions is None:
        extensions = [".mp4", ".mkv", ".avi", ".mov"]

    results = []
    for root, _dirs, files in library_path.walk():
        for filename in files:
            file = root / filename
            if file.suffix in extensions:
                print(f"Inspecting '{filename}'...")
                video_tracks = get_video_tracks(file)
                audio_tracks = get_audio_tracks(file)
                subtitles = get_subtitles(file)
                results.append(
                    {
                        "filename": str(file.relative_to(library_path)),
                        "container": get_format(file),
                        "video_tracks": make_track_entries(video_tracks),
                        "audio_tracks": make_track_entries(audio_tracks),
                        "subtitle_tracks": make_track_entries(subtitles),
                    }
                )

    return results


def save_json(results: list[dict], path: Path) -> None:
    """Save results to a JSON file."""
    with open(path, "w", encoding="utf-8") as file:
        json.dump(results, file, indent=2, ensure_ascii=False)

    print(f"Saved JSON: {path}")


def save_csv(results: list[dict], path: Path) -> None:
    """Save results to a CSV file."""
    with open(path, "w", newline="", encoding="utf-8") as file:
        writer = csv.writer(file)
        writer.writerow(["filename", "type", "index", "language", "codec", "title"])
        for entry in results:
            for track in entry["audio_tracks"]:
                writer.writerow(
                    [
                        entry["filename"],
                        "audio",
                        track["index"],
                        track["language"],
                        track["codec"],
                        track["title"],
                    ]
                )
            for track in entry["subtitle_tracks"]:
                writer.writerow(
                    [
                        entry["filename"],
                        "subtitle",
                        track["index"],
                        track["language"],
                        track["codec"],
                        track["title"],
                    ]
                )

    print(f"Save CSV: {path}")


def print_to_terminal(results: list[dict]) -> None:
    """Print results to terminal."""
    for entry in results:
        print(entry["filename"])
        show_track_entries(entry["audio_tracks"], name="audio track")
        show_track_entries(entry["subtitle_tracks"], name="subtitle track")


def scan_mode(
    library: Path,
    extensions: list[str],
    show: bool = True,
    json_file: Path | None = None,
    csv_file: Path | None = None,
) -> bool:
    """Run the scan and save/show the output."""
    if not (show or json_file or csv_file):
        print("No output specified. Use --print, --json or --csv.")
        return False

    scan_result = scan_video_library(library, extensions=extensions)

    if show:
        print_to_terminal(scan_result)

    if json_file:
        save_json(scan_result, json_file)

    if csv_file:
        save_csv(scan_result, csv_file)

    return True



"""Tests for srt_tools."""

import vidmux.srt_tools as srt_tools


def test_timestamp_conversion() -> None:
    """Test the conversion from timestamps to milliseconds and vice versa."""
    org_timestamp = "01:08:57,200"
    org_milliseconds = 4137200

    assert (
        srt_tools.timestamp_to_milliseconds(*org_timestamp.replace(",", ":").split(":"))
        == org_milliseconds
    )
    assert srt_tools.timestamp_from_milliseconds(org_milliseconds) == org_timestamp

    # Also test timestamp formats that differ slightly
    org_timestamp = "1:08:57,200".replace(",", ":").split(":")
    assert srt_tools.timestamp_to_milliseconds(*org_timestamp) == org_milliseconds

    org_timestamp = "01:08:57,2".replace(",", ":").split(":")
    assert srt_tools.timestamp_to_milliseconds(*org_timestamp) == org_milliseconds


def test_text_processing() -> None:
    """Test the correct shifting of all timestamps in a text."""
    # We also test the flexibility of the Regex search, one timestamp has ".TT" instead
    # of ",TTT" for milliseconds. Processing the text should also correct this.
    org_text = (
        "1\n"
        "01:08:57,200 --> 01:08:59,900\n"
        "Some text...\n\n"
        "2\n"
        "01:09:59.00 --> 01:10:00,200\n"
        "Some other text...\n\n"
    )

    shift_seconds = 0.9

    target_text = (
        "1\n"
        "01:08:58,100 --> 01:09:00,800\n"
        "Some text...\n\n"
        "2\n"
        "01:09:59,900 --> 01:10:01,100\n"
        "Some other text...\n\n"
    )

    shifted_text, count = srt_tools.process_text(org_text, shift_seconds)

    assert shifted_text == target_text
    assert count == 2



"""Tools for video inspection (format, streams, resolution, ...)."""

import json
import subprocess
from pathlib import Path


def get_format(file_path: Path) -> str:
    """Return the container format of a file (e.g., 'mov,mp4,m4a,3gp,3g2,mj2')."""
    command = [
        "ffprobe",
        "-v",
        "quiet",
        "-print_format",
        "json",
        "-show_format",
        str(file_path),
    ]
    result = subprocess.run(command, capture_output=True, text=True)
    try:
        info = json.loads(result.stdout)
        return info.get("format", {}).get("format_name", "")
    except Exception:
        return ""


def get_streams(file_path: Path, stream_type: str):
    """
    Extract media streams using ffprobe (ffmpeg).

    stream_type: "a" = Audio, "v" = Video, "s" = Subtitle
    """
    if stream_type not in ("a", "v", "s"):
        raise ValueError("Invalid stream_type. Must be 'a', 'v' or 's'.")

    cmd = [
        "ffprobe",
        "-v",
        "quiet",
        "-print_format",
        "json",
        "-show_streams",
        "-select_streams",
        stream_type,
        str(file_path),
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    try:
        info = json.loads(result.stdout)
        return info.get("streams", [])
    except json.JSONDecodeError:
        return []


def get_audio_tracks(file_path: Path):
    """Extract audio tracks using ffprobe (ffmpeg)."""
    return get_streams(file_path, "a")


def get_subtitles(file_path: Path):
    """Extract subtitles using ffprobe (ffmpeg)."""
    return get_streams(file_path, "s")


def get_video_tracks(file_path: Path):
    """Extract video tracks using ffprobe (ffmpeg)."""
    return get_streams(file_path, "v")


def get_video_resolution(path_or_url: str | Path) -> tuple[int, int] | None:
    """Get the video resolution for a file or URL."""
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-show_entries",
        "stream=width,height",
        "-of",
        "csv=p=0:s=x",
        str(path_or_url),
    ]
    try:
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT).decode().strip()
        width, height = output.split("x")
        return int(width), int(height)
    except subprocess.CalledProcessError as err:
        print("ffprobe error:", err.output.decode())
        return None



"""Define the rules used in the validation process."""

import re
from pathlib import Path

from vidmux.library_structure.core import registry, IssueCode, ValidationIssue, Severity


def normalize_name(name: str) -> str:
    """Normalize a folder/filename, i.e. delete (YYYY) and allowed suffixes."""
    # Delete '(YYYY)'
    name = re.sub(r"\(\d{4}\)", "", name)
    # Delete '[...]' (without deleting text between two pairs of brackets)
    name = re.sub(r"\[.*?\]", "", name).lower()

    return name.strip()


def is_file_in_separate_folder(path: Path) -> bool:
    """
    Check whether the file is in a separate folder with similar name.

    The folder name and filename can differ to a certain extent.

    Note: Currently this function is called twice per file. If this increases, consider
    using functools.lru_cache to save the result after the first call.
    """
    folder_name = normalize_name(path.parent.name)
    file_stem = normalize_name(path.stem)

    return folder_name in file_stem


@registry.register(default_severity=Severity.WARNING)
def check_file_in_own_folder(path: Path, params) -> list[ValidationIssue]:
    """Check whether the file is in its own subfolder."""
    issues = []

    if not is_file_in_separate_folder(path):
        issues.append(
            ValidationIssue(
                path=path,
                code=IssueCode.FILE_NOT_IN_FOLDER,
                message=f"File not in own folder ({path.parent.name})",
                severity=Severity.WARNING,
            )
        )

    return issues


@registry.register(
    default_severity=Severity.WARNING,
    default_params={
        "allowed_suffixes": r"(\s*[-–]\s*[A-Za-z0-9 ]+|\s*\[[A-Za-z0-9+\- ]+\])$"
    },
)
def check_filename_matches_folder(path: Path, params) -> list[ValidationIssue]:
    """Check whether file and folder name match."""
    issues = []

    # If the file is not in a separate folder this rule does not apply at all
    if not is_file_in_separate_folder(path):
        return issues

    folder_name = path.parent.name
    file_stem = path.stem

    # Strip allowed suffixes from filename, e.g. "- Director's Cut", "[EN+DE]"
    allowed_suffixes = re.compile(params["allowed_suffixes"])
    cleaned_file_stem = allowed_suffixes.sub("", file_stem).strip()
    cleaned_folder_name = allowed_suffixes.sub("", folder_name).strip()

    if cleaned_folder_name not in cleaned_file_stem:
        issues.append(
            ValidationIssue(
                path=path,
                code=IssueCode.FILE_AND_FOLDER_NAME_DIFFER,
                message=(
                    f"Filename '{file_stem}' differs from folder '{folder_name}' "
                    f"(except for allowed suffixes)"
                ),
                severity=Severity.WARNING,
            )
        )

    return issues


@registry.register(
    default_severity=Severity.ERROR,
    default_params={"bad_chars": r'<>:"/\\|?*'},
)
def check_for_bad_characters(path: Path, params) -> list[ValidationIssue]:
    """Check whether the filename contains illegal characters."""
    bad_chars = set(params["bad_chars"])
    found = sorted({char for char in path.name if char in bad_chars})
    if not found:
        return []

    return [
        ValidationIssue(
            path=path,
            code=IssueCode.BAD_CHARACTER_IN_NAME,
            message=f"Filename contains illegal character(s): {' '.join(found)}",
            severity=Severity.ERROR,
        )
    ]


@registry.register(default_severity=Severity.WARNING)
def check_year_in_filename(path: Path, params) -> list[ValidationIssue]:
    """Check whether a year (YYYY) is given in the filename."""
    issues = []
    if not re.search(r"\(\d{4}\)", path.stem):
        issues.append(
            ValidationIssue(
                path=path,
                code=IssueCode.MISSING_YEAR,
                message="No year in parentheses",
                severity=Severity.WARNING,
            )
        )

    return issues



"""Tools to work with the language field of audio tracks."""

import shlex
import subprocess
import sys
from pathlib import Path

from vidmux.video_inspection import get_audio_tracks, get_subtitles


def build_ffmpeg_language_command(
    input_file: Path,
    output_file: Path,
    audio_languages: list[str],
    subtitle_languages: list[str],
    audio_titles: list[str] = None,
    subtitle_titles: list[str] = None,
) -> list[str]:
    """Build command to set track languages using ffmpeg."""
    command = ["ffmpeg", "-i", str(input_file), "-map", "0", "-c", "copy"]

    for index, language in enumerate(audio_languages):
        command += [f"-metadata:s:a:{index}", f"language={language}"]
        if audio_titles and index < len(audio_titles):
            command += [f"-metadata:s:a:{index}", f"title={audio_titles[index]}"]

    for index, language in enumerate(subtitle_languages):
        command += [f"-metadata:s:s:{index}", f"language={language}"]
        if subtitle_titles and index < len(subtitle_titles):
            command += [f"-metadata:s:s:{index}", f"title={subtitle_titles[index]}"]

    command.append(str(output_file))

    return command


def set_languages(
    input_file: Path,
    output_file: Path,
    audio_languages: list[str],
    subtitle_languages: list[str],
    audio_titles: list[str] = None,
    subtitle_titles: list[str] = None,
    dry_run: bool = False,
):
    audio_streams = get_audio_tracks(input_file)
    subtitle_streams = get_subtitles(input_file)

    if len(audio_languages) != len(audio_streams):
        msg = (
            f"Error: Number of audio languages ({len(audio_languages)}) does not match "
            f"number of audio tracks ({len(audio_streams)})."
        )
        print(msg, file=sys.stderr)
        sys.exit(1)

    if subtitle_languages and len(subtitle_languages) != len(subtitle_streams):
        msg = (
            f"Error: Number of subtitle languages ({len(subtitle_languages)}) does not "
            f"match number of subtitle tracks ({len(subtitle_streams)})."
        )
        print(msg, file=sys.stderr)
        sys.exit(1)

    command = build_ffmpeg_language_command(
        input_file,
        output_file,
        audio_languages,
        subtitle_languages,
        audio_titles,
        subtitle_titles,
    )

    print("FFmpeg command:")
    print(" ".join(shlex.quote(part) for part in command))

    if not dry_run:
        subprocess.run(command)
